import enum
import datetime
from typing import Iterable, Optional


def str_to_date(date: str) -> datetime.date:
    if '-' in date:
        return datetime.datetime.strptime(date, '%Y-%m-%d').date()
    else:
        return datetime.datetime.strptime(date, '%Y%m%d').date()


def enum_lookup(value: str, e: Iterable[enum.Enum]) -> Optional[enum.Enum]:
    for member in list(e):
        if (value == member.name or value == member.value):
            return member
    return None



import datetime
import dataclasses
import re
import pathlib
import json
from typing import (List,
                    Dict,
                    Optional,
                    Iterator,
                    Any,
                    Sequence,
                    Union)
from loguru import logger
from . import utilities
from .client import Client, Classification, Site
from .episode import Episode
from . import episode
from .study import Study
from .series import Series
from .image import Image
from .mark import (
    BenignClassification,
    Conspicuity,
    MassClassification,
    BoundingBox,
    Mark
)

from .events import (
    Event,
    Events,
    BreastScreeningData,
    Screening,
    BaseEvent,
    Opinion,
    SideOpinion
)


class DB():
    """
    OMI-DB parser

    :param root_dir: Root directory of the OMI-DB. The `data` and `images`
        directories should sit somewhere below the root, and can be nested under
        sub-directories, e.g. ``data/sample/omi-db/(data|images)`` is fine.
    :param ignore_missing_images: If ``True``, the existence of dicom images
        belonging to a series will not be checked: parsing is based entirely on the
        JSON representations of the DICOM headers. Set to ``False`` to parse only
        those images for which you have *both* JSON and DICOM files for.
    :param clients: Only parse these clients, if they exist
    :param exclude_clients: Exclude these clients, even if they are in ``clients``
    """

    def __init__(self,
                 root_dir: Union[str, pathlib.Path],
                 ignore_missing_images: bool = True,
                 clients: Optional[Sequence[str]] = None,
                 exclude_clients: Optional[Sequence[str]] = None
                 ):

        root_dir = pathlib.Path(root_dir)
        self.ignore_missing_images = ignore_missing_images

        # The parent of the DATA/IMAGES directory
        self.real_root_dir = pathlib.Path()

        match = None
        for dir in root_dir.glob('**/'):
            match = re.match(r"(.*)/(data|images)", str(dir), re.I)
            if match:
                self.real_root_dir = pathlib.Path(match.group(1))
                break

        if match is None:
            raise ValueError('Failed to parse the given directory')

        if not clients:

            path_clients = (
                list(self.real_root_dir.glob('data/*')) +
                list(self.real_root_dir.glob('DATA/*'))
            )

            clients = [_.name for _ in path_clients]

        self.clients = set(clients)

        if exclude_clients:
            self.clients = set(clients) - set(exclude_clients)

    def __iter__(self) -> Iterator[Client]:
        """
        Iterates over all parsable clients found in the OMI-DB directory.

        :return: client_it: A :class:`omidb.client.Client` iterator
        """
        for client in self.clients:
            _data_dir = self._data_dir(client)
            studies: List[str] = []
            for study in _data_dir.glob('*/**'):

                # Extract study ID from path
                match = re.search(r"(\w+\d+)/(.*)", str(study))
                if match:
                    studies.append(match.group(2))

            try:
                episodes = self._episodes_from_imagedb(client, studies)

                classification: Classification = utilities.enum_lookup(  # type: ignore
                    self._nbss(client)['Classification'],
                    Classification
                )

                site: Site = utilities.enum_lookup(  # type: ignore
                    self._imagedb(client)['Site'],
                    Site
                )

                yield Client(
                    id=client,
                    episodes=episodes,
                    classification=classification,
                    site=site,
                )

            except Exception as e:
                logger.exception(
                    f'omidb: Failed to parse {client}, '
                )
                continue

    def _data_dir(self, client_id: str) -> pathlib.Path:
        """
        Path of the data directory corresponding to the client with ID
        `client_id`
        """
        p1 = self.real_root_dir / 'data' / client_id

        if p1.exists():
            return p1
        else:
            return self.real_root_dir / 'DATA' / client_id

    def _image_dir(self, client_id: str) -> pathlib.Path:
        """Path of the image directory corresponding to the client with ID
        `client_id`
        """
        p1 = self.real_root_dir / 'images' / client_id

        if p1.exists():
            return p1
        else:
            return self.real_root_dir / 'IMAGES' / client_id

    def _nbss_path(self, client_id: str) -> pathlib.Path:
        """Path of the NBSS json file corresponding to the client with ID
        `client_id`
        """

        p1 = self._data_dir(client_id) / ('nbss_' + client_id + '.json')

        if not p1.exists():
            p1 = self._data_dir(client_id) / ('NBSS_' + client_id + '.json')

        return p1

    def _nbss(self, client_id: str) -> Dict[str, Any]:
        """NBSS data corresponding to the client with ID `client_id`"""

        with open(self._nbss_path(client_id)) as f:
            nbss: Dict[str, Any] = json.load(f)
        return nbss

    def _imagedb(self, client_id: str) -> Dict[str, Any]:
        """IMAGEDB data corresponding to the client with ID `client_id`"""

        p1 = self._data_dir(client_id) / ('imagedb_' + client_id + '.json')

        if not p1.exists():
            p1 = self._data_dir(client_id) / ('IMAGEDB_' + client_id + '.json')

        with open(p1) as f:
            imagedb: Dict[str, Any] = json.load(f)

        return imagedb

    def _has_studies(self, client_id: str) -> bool:
        imagedb = self._imagedb(client_id)

        if (
            ('STUDIES' not in imagedb) or
            (not isinstance(imagedb['STUDIES'], dict))
        ):
            logger.error(
                f'No studies listed in IMAGEDB for client {client_id}'
            )

            return False

        return True

    def _parse_mark(self, mark_data: Dict[str, Any]) -> Mark:

        args: Dict[str, Any] = {}
        for param_name, key in zip((
            'architectural_distortion',
            'dystrophic_calcification',
            'fat_necrosis',
            'focal_asymmetry',
            'mass',
            'suspicious_calcifications',
            'milk_of_calcium',
            'other_benign_cluster',
            'plasma_cell_mastitis',
            'benign_skin_feature',
            'calcifications',
            'suture_calcification',
            'vascular_feature',
        ),
            (
            'ArchitecturalDistortion',
            'Dystrophic',
            'FatNecrosis',
            'FocalAsymmetry',
            'Mass',
            'SuspiciousCalcifications',
            'MilkOfCalcium',
            'OtherBenignCluster',
            'PlasmaCellMastitis',
            'Skin',
            'WithCalcification',
            'SutureCalcification',
            'Vascular',
        )):

            args[param_name] = True if mark_data.get(key) else None

        args['benign_classification'] = (
            utilities.enum_lookup(
                str(mark_data.get('BenignClassification')),
                BenignClassification
            )
        )

        args['conspicuity'] = utilities.enum_lookup(
            str(mark_data.get('Conspicuity')),
            Conspicuity
        )

        args['mass_classification'] = (
            utilities.enum_lookup(
                str(mark_data.get('MassClassification')),
                MassClassification
            )
        )

        args['lesion_id'] = str(mark_data['LinkedNBSSLesionNumber'])

        args['id'] = str(mark_data['MarkID'])

        args['boundingBox'] = BoundingBox(
            x1=int(mark_data['X1']),
            y1=int(mark_data['Y1']),
            x2=int(mark_data['X2']),
            y2=int(mark_data['Y2']),
        )

        return Mark(**args)

    def _parse_series(self,
                      client_id: str,
                      study: str,
                      study_data: Dict[str, Any]) -> List[Series]:

        series_list = []
        for series, series_dic in study_data.items():

            if not isinstance(series_dic, dict):
                continue

            image_list = list(series_dic.keys())

            images = []
            for image in image_list:

                dcm_path = (
                    self._image_dir(client_id) /
                    study /
                    (image + '.dcm')
                )

                json_path = (
                    self._data_dir(client_id) /
                    study /
                    (image + '.json')
                )

                # Due to inconsistency in file naming
                if not json_path.exists():
                    json_path = pathlib.Path(
                        str(json_path).replace('.json', '.dcm.json')
                    )

                if not json_path.exists():
                    logger.error(
                        f'Image metadata {json_path} does not exist, skipping.'
                    )
                    continue

                # Skip this image if no dcm file
                if not self.ignore_missing_images:
                    if not dcm_path.is_file():
                        logger.error(
                            f'Image {dcm_path} of does not exist, skipping.'
                        )
                        continue

                marks: List[Mark] = []

                image_marks_data = series_dic.get(image)
                if isinstance(image_marks_data, dict):

                    for mark_id, mark_data in image_marks_data.items():
                        try:
                            marks.append(self._parse_mark(mark_data))
                        except:
                            logger.exception(
                                f"Failed to parse mark data of {client_id} "
                                f"{image}.dcm"
                            )
                            continue

                images.append(
                    Image(image,
                          dcm_path,
                          json_path,
                          marks
                          )
                )

            series_list.append(Series(id=series, images=images))

        return series_list

    def _parse_events(self, episode_data: Dict[str, Any]) -> Events:
        event_kwargs = {
            'screening': episode_data.get('SCREENING'),
            'assessment': episode_data.get('ASSESSMENT'),
            'biopsy_wide': episode_data.get('BIOPSYWIDE'),
            'biopsy_fine': episode_data.get('BIOPSYFINE'),
            'clinical': episode_data.get('CLINICAL'),
            'surgery': episode_data.get('SURGERY'),
        }

        for key, value in event_kwargs.items():

            if not value:
                continue

            if key == 'screening':
                event_kwargs['screening'] = self._parse_screening_event(
                    value
                )
            else:
                event_kwargs[key] = self._parse_base_event(
                    value
                )

        return Events(**event_kwargs)

    def _parse_base_event(self, data: Dict[str, Any]) -> BaseEvent:

        dates: List[datetime.date] = []

        for i, side in enumerate(['L', 'R']):

            side_data = data.get(side)

            if not side_data:
                continue

            for lesion_id, lesion in side_data.items():
                if lesion.get('DatePerformed'):
                    date = utilities.str_to_date(lesion['DatePerformed'])
                    if date not in dates:
                        dates.append(date)

        left_opinion: SideOpinion = utilities.enum_lookup(  # type: ignore
            str(data.get('left_opinion')),
            SideOpinion
        )

        right_opinion: SideOpinion = utilities.enum_lookup(  # type: ignore
            str(data.get('right_opinion')),
            SideOpinion
        )

        return BaseEvent(
            left_opinion=left_opinion,
            right_opinion=right_opinion,
            dates=dates,
        )

    def _parse_screening_event(self, data: Dict[str, Any]) -> Screening:

        breast_data: List[Optional[BreastScreeningData]] = [None, None]
        dates: List[datetime.date] = []

        for i, side in enumerate(['L', 'R']):

            side_data = data.get(side)

            if not side_data:
                continue

            date: Optional[datetime.date] = None
            if side_data.get('DateTaken'):
                date = utilities.str_to_date(side_data['DateTaken'])
                if date not in dates:
                    dates.append(date)

            opinion: Optional[Opinion] = utilities.enum_lookup(  # type: ignore
                side_data.get('Opinion'),
                Opinion
            )

            breast_data[i] = (
                BreastScreeningData(
                    date=date,
                    equipment_make_model=side_data.get('EquipmentMakeModel'),
                    opinion=opinion,
                )
            )

        left_opinion: Optional[SideOpinion] = utilities.enum_lookup(  # type: ignore
            str(data.get('left_opinion')),
            SideOpinion
        )

        right_opinion: Optional[SideOpinion] = utilities.enum_lookup(  # type: ignore
            str(data.get('right_opinion')),
            SideOpinion
        )

        return Screening(
            left_opinion=left_opinion,
            right_opinion=right_opinion,
            left=breast_data[0],
            right=breast_data[1],
            dates=dates,
        )

    def _episodes_from_imagedb(
        self,
        client_id: str,
        studies: Optional[List[str]]) -> List[Episode]:
        """The beast, the animal, the dark knight.

        Prepare and instantiate all objects required for this client and the
        associated studies. Models are generated for only those images that
        exist locally.

        Note studies that cannot be linked to an NBSS event will not be
        aggregated.
        """

        if not self._has_studies(client_id):
            return []

        nbss = self._nbss(client_id)
        imagedb = self._imagedb(client_id)

        # Events for each episode
        events: Dict[str, Events] = {}
        for episode_id, nbss_episode in nbss.items():
            if isinstance(nbss_episode, dict):
                events[episode_id] = self._parse_events(nbss_episode)

        studies = [] if studies is None else studies
        # Studies for each episode
        episodes: Dict[str, List[Study]] = {}
        for study_id, study_data in imagedb['STUDIES'].items():

            if study_id not in studies:  # These exist locally
                if not study_id:
                    logger.error(
                        f'Empty study listed in IMAGEDB for client '
                        f'{client_id}...skipping'
                    )

                else:

                    logger.error(
                        f'{client_id}/{study_id} directory does not exist '
                        'but listed in IMAGEDB...skipping'
                    )

                continue

            series_list = self._parse_series(client_id,
                                             study_id,
                                             study_data)

            study_date = (
                utilities.str_to_date(study_data['StudyDate'])
                if study_data.get('StudyDate') else None
            )

            if study_data is None:
                logger.warning(f'{client_id}/{study_id} has no study date')
            # Try to extract episode ID by study-date <->event-date
            elif (
                ('EpisodeID' not in study_data) or
                (not study_data['EpisodeID']) or
                (study_data['EpisodeID'] not in nbss)
            ):
                logger.warning(
                    f"Episode {study_data.get('EpisodeID')} not found in NBSS "
                    f"for {client_id}/{study_id}, "
                    f"attempting link via event dates"
                )

                for episode_id, episode_events in events.items():
                    for field in dataclasses.fields(episode_events):
                        event = getattr(episode_events, field.name)
                        if event and (study_date in event.dates):
                            study_data['EpisodeID'] = episode_id
                            logger.info(
                                f'Linked study {study_id} to episode '
                                f'{episode_id}'
                            )
                            break

            # If still not episode ID, skip
            if (
                ('EpisodeID' not in study_data) or
                (not study_data['EpisodeID'])
            ):

                logger.error(
                    f'EpisodeID not found for {client_id}/{study_id}, skipping'
                )

                continue

            # Report if there are no events match study date
            matched_events = []
            if (study_data['EpisodeID'] in events) or (not study_data['EpisodeID']):

                episode_events = events[study_data['EpisodeID']]

                for field in dataclasses.fields(episode_events):
                    event = getattr(episode_events, field.name)
                    if event and (study_date in event.dates):
                        matched_events.append(getattr(Event, field.name))

                if not matched_events:
                    logger.warning(
                        f"No events in episode "
                        f"{study_data['EpisodeID']} match study date of "
                        f"{client_id}/{study_id}"
                    )

            else:
                logger.warning(
                    f"Episode {study_data['EpisodeID']} has no events "
                    "(episode not found in NBSS)"
                )
                continue

            # Now add studies to the episode
            study = Study(
                id=study_id,
                series=series_list,
                date=study_date,
                event_type=matched_events)

            if study_data['EpisodeID'] not in episodes:
                episodes[study_data['EpisodeID']] = [study]
            else:
                episodes[study_data['EpisodeID']].append(study)

        # Complete episode
        out = []
        for episode_id, episode_studies in episodes.items():

            if (episode_id not in nbss) or (episode_id not in events):

                logger.error(
                    f'Episode "{episode_id}" for client {client_id} '
                    'does not exist in NBSS, or has no events?'
                    'This episode is listed in IMAGEDB...skipping'
                )

                continue

            nbss_episode = nbss[episode_id]
            is_closed = (
                True if nbss_episode.get('EpisodeIsClosed') == 'Y'
                else False
            )

            ep_type: Optional[episode.Type] = utilities.enum_lookup(  # type: ignore
                nbss_episode.get('EpisodeType'),
                episode.Type
            )

            ep_action: Optional[episode.Action] = utilities.enum_lookup(  # type: ignore
                nbss_episode.get('EpisodeAction'),
                episode.Action
            )

            out.append(
                Episode(
                    id=episode_id,
                    events=events[episode_id],
                    studies=episode_studies,
                    type=ep_type,
                    action=ep_action,
                    opened_date=nbss_episode.get('EpisodeOpenedDate'),
                    closed_date=nbss_episode.get('EpisodeClosedDate'),
                    is_closed=is_closed,
                )
            )

        return out



from .client import Client, Classification
from .episode import Episode
from .events import Event, SideOpinion
import copy
from typing import List

def episode_is_malignant(episode: Episode) -> bool:
    """
    Returns ``True`` if ``episode`` contains either a
    ``episode.events.surgery`` or ``episode.events.biopsy_wide`` event where
    the left or right opinion is malignant (``omidb.events.SideOpinion.OM``).

    :param episode: The episode to evaluate
    """

    try:

        surgery = episode.events.surgery
        return (
            surgery.left_opinion == SideOpinion.OM or  # type: ignore
            surgery.right_opinion == SideOpinion.OM  # type: ignore
        )
    except AttributeError as e:
        pass

    try: # What about biopsy fine?
        biopsy = episode.events.biopsy_wide
        return (
            biopsy.left_opinion == SideOpinion.OM or  # type: ignore
            biopsy.right_opinion == SideOpinion.OM  # type: ignore
        )
    except AttributeError:
        pass

    return False


def has_prior(client: Client) -> bool:
    """
    Returns ``True`` if ``client`` has a non-malignant episode earlier than a
    malignant episode; ``False`` otherwise.

    :param client: A classified client with more than one episode
    """

    if client.classification != Classification.M:
        return False

    nm_date = None  # earliest non-malignant date
    m_date = None  # most recent malignant date

    for episode in client.episodes:
        if None in (episode.id, episode.opened_date):
            continue

        date = episode.opened_date

        if episode_is_malignant(episode):
            if (m_date is None) or (date > m_date):
                m_date = date
        else:
            if (nm_date is None) or (date < nm_date):
                nm_date = date

        if nm_date and m_date:
            if nm_date < m_date:
                return True

    return False


def filter_studies_by_event_type(
    client: Client,
    event_type: List[Event],
    exact_match: bool = True
) -> Client:
    """
    Remove studies (``omidb.study.Study``) from ``client`` whose event type
    includes one or all (if ``exact_match`` is ``True``) of those listed by
    ``event_type``.

    :param client: A classified client with more than one episode
    :param event_type: A list of event types
    :param exact_match: If ``True`` the event types of a study must all of
        those in ``event_type``.
    """

    client2 = copy.deepcopy(client)

    for idx, episode in enumerate(client.episodes):
        for study in episode.studies:
            if study.event_type:
                if exact_match and set(study.event_type) != set(event_type):
                    client2.episodes[idx].studies.remove(study)
                elif not any([_ in study.event_type for _ in event_type]):
                    client2.episodes[idx].studies.remove(study)

    return client2



from .parser import DB
from . import image
from . import mark
from . import series
from . import study
from . import episode
from . import client
from . import utilities
from . import filters
from . import classificationtools
from . import commands
from loguru import logger

logger.disable("omidb")



import json
import pathlib
import pydicom
import matplotlib
import matplotlib.pyplot as plt
import datetime
from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Union, Tuple
from .events import Events
from .mark import Mark


@dataclass
class Image:
    """
    Container for a mammogram, stored in the DICOM_ format. An image can have
    zero or more marks for one or many lesions.

    :param id: SOP Instance UID, a unique identifier
    :param dcm_path: Path to the dicom image
    :param json_path: Path to the JSON file storing DICOM metadata
    :param marks: A list of marks or annotations, represented by
        :class:`omidb.mark.Mark`

    .. _DICOM: https://www.dicomstandard.org/
    """

    id: str
    dcm_path: pathlib.Path
    json_path: pathlib.Path
    marks: Optional[List[Mark]]
    _dcm: Optional[pydicom.FileDataset] = None
    _json: Optional[Dict[str, Any]] = None

    @property
    def dcm(self) -> pydicom.FileDataset:
        """
        Returns a :class:`pydicom.dataset.FileDataset`, representing a parsed DICOM file
        """

        if not self._dcm:
            self._dcm = pydicom.dcmread(str(self.dcm_path))

        return self._dcm

    def dcm_tags(self, tags: Union[str, List[str]]) -> pydicom.FileDataset:
        """
        Returns a pydicom.dataset.FileDataset, with only the specified tags of
        a parsed DICOM file and no pixel data
        """

        if isinstance(tags, str):
            tags = [tags]

        return pydicom.dcmread(str(self.dcm_path),
                               stop_before_pixels=True,
                               specific_tags=tags)

    @property
    def attributes(self) -> Optional[Dict[str, Any]]:
        """
        Access DICOM metadata via the JSON representation
        """

        if not self._json:
            with open(self.json_path) as f:
                self._json = json.load(f)

        return self._json

    def plot(
        self,
        ax: Optional[matplotlib.axes.Axes] = None
    ) -> matplotlib.image.AxesImage:
        """
        Plot the dicom
        """

        if not ax:
            fig, ax = plt.subplots()

        return ax.imshow(self.dcm.pixel_array, cmap=plt.cm.bone)



import datetime
from dataclasses import dataclass
from typing import Optional, List
import enum
from .events import Events
from .study import Study
from .lesion import Lesion


@enum.unique
class Type(enum.Enum):
    """Type of episode"""
    CA = 'Continued Assessment'
    CD = 'Delayed Treatment'
    CF = 'Follow-up after treatment'
    CI = 'Interval case'
    CR = 'Local Recurrence'
    F = 'First Call'
    G = 'GP Referral'
    H = 'Higher Risk'
    N = 'Non-rout Recall'
    R = 'Routine Recall'
    S = 'Self Referral'
    X = 'Other'


@enum.unique
class Action(enum.Enum):
    """Episode action"""
    EC = 'Early Recall for Clinic'
    ES = 'Early Recall for Screening'
    FN = 'Fine Needle Aspiration'
    FP = 'Follow-up (Post-treatment)'
    FV = 'Further X-ray views'
    IP = 'Inpatient biopsy'
    MT = 'Medical Treatment'
    NA = 'No Action from this procedure'
    R2 = 'Routine second film opinion (obsolete)'
    RC = 'Review in clinic'
    RF = 'Referral to consultant/GP'
    RR = 'Routine recall for screening'
    ST = 'Surgical Treatment'
    TR = 'Repeat Film (technical)'
    WB = 'Wide Bore Needle'


@dataclass
class Episode:
    """
    An episode contains a set of medical procedures or events associated with
    the treatment or diagnosis of a clinical condition. Medical imaging studies
    are included with each episode, and are typically (but not always) linked
    to one or more events.

    :param id: NBSS episode identifier, only unique within client
    :param events: A set of medical procedures associated with the episode
    :param studies: List of :class:`omidb.study.Study` s where collections of
        screening and diagnostic images reside
    :param type: Enumeration defining the type of episode
    :param action: Enumeration defining the action outcome of the episode
    :param opened_date: Date that the episode opened
    :param closed_date: Date that the episode closed
    :param is_closed: boolean signifying whether the episode is closed
    :param lesions: A list of :class:`omidb.lesion.Lesion` s, examined in the
        episode
    """
    id: str
    events: Events
    studies: List[Study]
    type: Optional[Type] = None
    action: Optional[Action] = None
    opened_date: Optional[datetime.date] = None
    closed_date: Optional[datetime.date] = None
    is_closed: Optional[bool] = None
    lesions: Optional[List[Lesion]] = None



from dataclasses import dataclass
import enum
from typing import List
from .episode import Episode


@enum.unique
class Classification(enum.Enum):
    """The global classification of a client"""
    N = 'Normal'
    M = 'Malignant'
    B = 'Benign'
    CI = 'Interval Case'


@enum.unique
class Site(enum.Enum):
    """Site/centre where the medical examination took place"""
    adde = 'adde'
    jarv = 'jarv'
    stge = 'stge'


@dataclass
class Client:
    """
    A client represents a patient who has attended the NHS screening programme.

    Each client will have one or more :class:`omidb.episode.Episode` s, for
    which (anonymised and pseudonymised) NBSS information can be found.

    :param id: Client identifier, typically a four letter ID followed by a
        series of digits, e.g. `optm1`, `demd7050`
    :param episodes: A list of :class:`omidb.episode.Episode` s associated with
        the client
    :param classification: Enumeration specifying the global classification of
        the client. Note that the classification is not extracted from NBSS,
        but is based on logic applied to to the within-event opinions.
    :param site: Enumeration signifying the clinical site/centre where
        screening took place
    """

    id: str
    episodes: List[Episode]
    classification: Classification
    site: Site



import itertools
from typing import List, Set, Dict, Tuple, Optional, Union
from dataclasses import dataclass
from enum import Enum


transforms = {}
transforms['CystAspirated'] = lambda x: True if x == 'Y' else False

Side = Enum('Side', 'L R')

LesionCodeToLesionDescription = {
    'AS': 'Asymmetry',
    'CA': 'Calcification only',
    'CY': 'Cyst',
    'DS': 'Distortion',
    'LN': 'Lymph node',
    'MA': 'Mass',
    'MC': 'Mass with calcification',
    'NA': 'No significant abnormality',
    'ZZ': 'Clinical abnormality',
}

LesionDescriptionToLesionCode = {
    v: k for k, v in LesionCodeToLesionDescription.items()
}

LesionDescriptionCode = Enum(  # type: ignore
    'LesionDescriptionCode',
    list(LesionCodeToLesionDescription.keys())
)

LesionDescription = Enum(  # type: ignore
    'LesionDescription',
    list(LesionCodeToLesionDescription.values())
)


LesionPosition = Enum(  # type: ignore
    'LesionPosition',
    [
        ''.join(_) for _ in
        list(itertools.product('LR', 'ABCDE', '1234')) + ['LM', 'RM']
    ]
)


@dataclass
class Lesion:
    side: Side
    cyst_aspirated: Optional[bool] = None
    description: Optional[LesionDescriptionCode] = None
    position: Optional[LesionPosition] = None
    notes: Optional[str] = None



import enum
import itertools
from dataclasses import dataclass
import datetime
from typing import List, Set, Dict, Tuple, Optional, Any


@enum.unique
class Opinion(enum.Enum):
    """Opinion after a procedure"""
    A1 = 'Assess normal'
    A2 = 'Assess benign'
    A3 = "Assess uncert'n"
    A4 = 'Assess suspic'
    A5 = 'Assess malig'
    B1 = 'Unsatis/Normal'
    B2 = 'Benign'
    B3 = 'Benign unc mal'
    B4 = 'Susp of malig'
    B5 = 'Malignant'
    BA = 'Clinical'
    C1 = 'Cyt unsatis'
    C2 = 'Cyt benign'
    C3 = 'Cyt atypia'
    C4 = 'Cyt susp malig'
    C5 = 'Cyt malig'
    H0 = 'Hist unreported'
    H1 = 'Histol normal'
    H2 = 'Histol benign'
    H5 = 'Hist malignant'
    I1 = 'Imaging normal'
    I2 = 'Imaging benign'
    I3 = 'Imaging uncertain'
    I4 = 'Imaging suspicious'
    I5 = 'Imaging malig'
    MRI1 = 'MRI normal'
    MRI2 = 'MRI benign'
    MRI3 = 'MRI indeterminate'
    MRI4 = 'MRI suspicious'
    MRI5 = 'MRI malig'
    OB = 'O Benign'
    OM = 'O Malignant'
    ON = 'O Normal'
    OS = 'O Suspicious'
    OU = 'O Uncertain'
    P1 = 'Clin normal'
    P2 = 'Clin benign'
    P3 = 'Clin uncertain'
    P4 = 'Clin suspicious'
    P5 = 'Clin malignant'
    R1 = 'Rad normal'
    R2 = 'Rad benign'
    R3 = 'Rad uncertain'
    R4 = 'Rad suspicious'
    R5 = 'Rad malig'
    RB = 'R Benign'
    RM = 'R Malignant'
    RN = 'R Normal'
    RO = 'R Unreported'
    RS = 'R Suspicious'
    RU = 'R Uncertain'
    SH = 'History suspic'
    U1 = 'USS normal'
    U2 = 'USS benign'
    U3 = 'USS uncertain'
    U4 = 'USS suspicious'
    U5 = 'USS malig'


@enum.unique
class SideOpinion(enum.Enum):
    """Side-specific opinion"""
    OB = 'Benign'
    OM = 'Malignant'
    ON = 'Normal'
    OS = 'Suspicious'
    OU = 'Uncertain'


@enum.unique
class Event(enum.Enum):
    """Type of event. Corresponds to the fields of :class:`omidb.events.Events`"""
    assessment = 'assessment'
    biopsy_fine = 'biopsy_fine'
    biopsy_wide = 'biopsy_wide'
    clinical = 'clinical'
    screening = 'screening'
    surgery = 'surgery'


@dataclass
class BaseEvent:
    """
    Generic NBSS information pertaining to a medical procedure.

    :param left_opinion: Code summarising the diagnosis of the left breast
    :param right_opinion: Code summarising the diagnosis of the right breast
    :param dates: Dates extracted from left and right lesion data
    """
    left_opinion: Optional[SideOpinion] = None
    right_opinion: Optional[SideOpinion] = None
    dates: Optional[List[datetime.date]] = None


@dataclass
class BreastScreeningData:
    """
    Side-specific breast-screening data.

    :param date: Date that the examination took place
    :param equipment_make_model: Short description of the equipment used for
        the examination. Typically the manufacturer of the imaging device.
    :param opinion: Enumeration representing the opinion code
    """
    date: Optional[datetime.date] = None
    equipment_make_model: Optional[str] = None
    opinion: Optional[Opinion] = None


@dataclass
class Screening(BaseEvent):
    """
    NBSS information pertaining to a screening procedure.

    :param left: Screening information for the left breast
    :param right: Screening information for the right breast
    """
    left: Optional[BreastScreeningData] = None
    right: Optional[BreastScreeningData] = None


@dataclass
class Events:
    """
    A collection of events, for a given :class:`omidb.core.Episode`.
    """
    screening: Optional[Screening] = None
    assessment: Optional[BaseEvent] = None
    clinical: Optional[BaseEvent] = None
    biopsy_wide: Optional[BaseEvent] = None
    biopsy_fine: Optional[BaseEvent] = None
    surgery: Optional[BaseEvent] = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omidb', 'omidb.commands']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'loguru>=0.4.1,<0.5.0',
 'matplotlib>=3.1.2,<4.0.0',
 'pydicom>=1.4.1,<2.0.0']

entry_points = \
{'console_scripts': ['omidb = omidb:commands.main']}

setup_kwargs = {
    'name': 'omidb',
    'version': '0.5.0',
    'description': 'Python package and CLI for The OPTIMAM Mammography Image Database (OMI-DB)',
    'long_description': None,
    'author': 'Dominic Ward',
    'author_email': 'dominic.ward1@nhs.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import List, Iterable, Tuple, Union, Any, no_type_check, Optional, Iterator
from collections import namedtuple
import click
import csv
from ..client import Client
from ..image import Image
from ..study import Study
from ..series import Series
from ..episode import Episode
from ..parser import DB
from loguru import logger


def flatten(clients: Iterable[Client]) -> Iterator[
        Tuple[Client,
              Episode,
              Study,
              Series,
              Image]]:
    for client in clients:
        for episode in client.episodes:
            for study in episode.studies:
                for series in study.series:
                    for image in series.images:
                        yield(client, episode, study, series, image)


DicomAttributes = namedtuple(
    'DicomAttributes',
    'Manufacturer PresentationIntentType ImageLaterality ViewPosition ViewModCodeValue ViewModCodeMeaning'
)


@no_type_check
def extract_dicom_attributes(image: Image) -> DicomAttributes:
    """ Forcefully extract dicom attributes """
    try:
        manufacturer = image.attributes['00080070']['Value'][0]
    except:
        manufacturer = None

    try:
        presentation_intent_type = image.attributes['00080068']['Value'][0]
    except:
        presentation_intent_type = None

    try:
        image_laterality = image.attributes['00200062']['Value'][0]
    except:
        image_laterality = None

    try:
        view_position = image.attributes['00185101']['Value'][0]
    except:
        view_position = None

    try:
        view_mod_code_value = (
            image.attributes['00540220']['Value'][0]
            ['00080100']['Value'][0]
        )
    except:
        view_mod_code_value = None

    try:
        view_mod_code_meaning = (
            image.attributes['00540220']['Value'][0]
            ['00080104']['Value'][0]
        )
    except:
        view_mod_code_meaning = None

    return DicomAttributes(
        manufacturer,
        presentation_intent_type,
        image_laterality,
        view_position,
        view_mod_code_value,
        view_mod_code_meaning
    )


class DataWriter:
    def __init__(self) -> None:

        # Set up header
        self.rows: List[List[Any]] = [[
            'ClientID',
            'ClientClassification',
            'Site',
            'EpisodeID',
            'EpisodeType',
            'EpisodeAction',
            'EpisodeOpenedDate',
            'EpisodeClosedDate',
            'StudyInstanceUID',
            'StudyDate',
            'EventType',
            'SeriesInstanceUID',
            'SOPInstanceUID',
            'NumberOfMarks',
        ]]

        self.rows[0] += list(DicomAttributes._fields)

    def add(self,
            client: Client,
            episode: Episode,
            study: Study,
            series: Series,
            image: Image) -> None:

        tags = extract_dicom_attributes(image)

        event_types: List[Union[None, enums.Event]] = [None]

        if study.event_type:
            event_types = [e.name for e in study.event_type]

        num_marks: int = len(image.marks) if image.marks else 0

        # 1 row per event
        for event_type in event_types:

            ep_action = episode.action.name if episode.action else None
            ep_type = episode.type.name if episode.type else None

            self.rows.append(
                [
                    client.id,
                    client.classification.name,
                    client.site.name,
                    episode.id,
                    ep_type,
                    ep_action,
                    episode.opened_date,
                    episode.closed_date,
                    study.id,
                    study.date,
                    event_type,
                    series.id,
                    image.id,
                    num_marks,
                    *tags,
                ]
            )

    def write(self, outfile: str) -> None:
        with open(outfile, 'w') as f:
            writer = csv.writer(f, delimiter=',')
            writer.writerows(self.rows)


def run(db_path: str,
        outfile: str,
        clients: Optional[str] = None) -> None:

    client_list = None
    if clients:
        with open(clients, 'r') as f:
            client_list = [_.strip() for _ in f.readlines()]

    db = DB(db_path, clients=client_list)
    writer = write_client_data(db)
    writer.write(outfile)


def write_client_data(
    clients: Iterable[Client],
) -> DataWriter:
    writer: DataWriter = DataWriter()
    for client, episode, study, series, image in flatten(clients):
        writer.add(client, episode, study, series, image)
    return writer


@click.command('summarise')
@click.argument('db', type=click.Path(exists=True))
@click.argument('output-file', type=click.Path(exists=False))
@click.option(
    '--clients-file',
    type=click.Path(exists=True),
    help='File containing a list of clients to parse'
)
@click.option(
    '--log-file',
    type=click.Path(exists=False),
    help='Log to this file'
)
def cli(db: str, output_file: str, clients_file: str, log_file: str) -> None:
    """"Write a csv file, OUTPUT_FILE, summarising the content of OMI-DB,
    located at DB
    """

    logger.enable('omidb')

    if log_file:
        logger.remove()
        logger.add(log_file)

    run(db, output_file, clients_file)
